/*  hw-data.c -- selected hardware specific data
 *  Copyright (C) 2004--2009  SEIKO EPSON CORPORATION
 *
 *  License: GPLv2+|iscan
 *  Authors: AVASYS CORPORATION
 *
 *  This file is part of the SANE backend distributed with Image Scan!
 *
 *  Image Scan!'s SANE backend is free software.
 *  You can redistribute it and/or modify it under the terms of the GNU
 *  General Public License as published by the Free Software Foundation;
 *  either version 2 of the License or at your option any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *  You ought to have received a copy of the GNU General Public License
 *  along with this package.  If not, see <http://www.gnu.org/licenses/>.
 *
 *
 *  Linking Image Scan!'s SANE backend statically or dynamically with
 *  other modules is making a combined work based on this SANE backend.
 *  Thus, the terms and conditions of the GNU General Public License
 *  cover the whole combination.
 *
 *  As a special exception, the copyright holders of Image Scan!'s SANE
 *  backend give you permission to link Image Scan!'s SANE backend with
 *  SANE frontends that communicate with Image Scan!'s SANE backend
 *  solely through the SANE Application Programming Interface,
 *  regardless of the license terms of these SANE frontends, and to
 *  copy and distribute the resulting combined work under terms of your
 *  choice, provided that every copy of the combined work is
 *  accompanied by a complete copy of the source code of Image Scan!'s
 *  SANE backend (the version of Image Scan!'s SANE backend used to
 *  produce the combined work), being distributed under the terms of
 *  the GNU General Public License plus this exception.  An independent
 *  module is a module which is not derived from or based on Image
 *  Scan!'s SANE backend.
 *
 *  As a special exception, the copyright holders of Image Scan!'s SANE
 *  backend give you permission to link Image Scan!'s SANE backend with
 *  independent modules that communicate with Image Scan!'s SANE
 *  backend solely through the "Interpreter" interface, regardless of
 *  the license terms of these independent modules, and to copy and
 *  distribute the resulting combined work under terms of your choice,
 *  provided that every copy of the combined work is accompanied by a
 *  complete copy of the source code of Image Scan!'s SANE backend (the
 *  version of Image Scan!'s SANE backend used to produce the combined
 *  work), being distributed under the terms of the GNU General Public
 *  License plus this exception.  An independent module is a module
 *  which is not derived from or based on Image Scan!'s SANE backend.
 *
 *  Note that people who make modified versions of Image Scan!'s SANE
 *  backend are not obligated to grant special exceptions for their
 *  modified versions; it is their choice whether to do so.  The GNU
 *  General Public License gives permission to release a modified
 *  version without this exception; this exception also makes it
 *  possible to release a modified version which carries forward this
 *  exception.
 */


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "hw-data.h"

#include <ctype.h>
#include <locale.h>
#include <stddef.h>
#include <string.h>
#include <time.h>


#include "command.h"
#include "profile.c"

struct ScannerName
{
  char *overseas;
  char *japan;
};
typedef struct ScannerName scanner_name_t;

struct ScannerData
{
  char *fw_name;
  int profile_ID;
  int command_ID;
  const scanner_name_t name;
};
typedef struct ScannerData scanner_data_t;

struct EpsonScanCommand
{
  int command_ID;
  unsigned char set_focus_position;
  unsigned char feed;
  unsigned char eject;
};
typedef struct EpsonScanCommand scan_command_t;

const struct ScannerData scanner_data[] = {
  {"GT-10000", 0x05, 0x05, {NULL, NULL}},
  {"ES-6000", 0x05, 0x05, {NULL, NULL}},
  {"Perfection610", 0x06, 0x01, {"Perfection 610", NULL}},
  {"GT-6600", 0x06, 0x01, {NULL, NULL}},
  {"Perfection1200", 0x07, 0x01, {"Perfection 1200", NULL}},
  {"GT-7600", 0x07, 0x01, {NULL, NULL}},
  {"Expression1600", 0x0D, 0x02, {"Expression 1600", NULL}},
  {"ES-2000", 0x0D, 0x02, {NULL, NULL}},
  {"Expression1640XL", 0x0F, 0x04, {"Expression 1640XL", NULL}},
  {"ES-8500", 0x0F, 0x04, {NULL, NULL}},
  {"Perfection640", 0x15, 0x01, {"Perfection 640", NULL}},
  {"GT-6700", 0x15, 0x01, {NULL, NULL}},
  {"Perfection1640", 0x16, 0x01, {"Perfection 1640", NULL}},
  {"GT-8700", 0x16, 0x01, {NULL, NULL}},
  {"Perfection1240", 0x18, 0x01, {"Perfection 1240", NULL}},
  {"GT-7700", 0x18, 0x01, {NULL, NULL}},
  {"GT-30000", 0x1A, 0x05, {NULL, NULL}},
  {"ES-9000H", 0x1A, 0x05, {NULL, NULL}},
  {"Expression1680", 0x1B, 0x02, {"Expression 1680", NULL}},
  {"ES-2200", 0x1B, 0x02, {NULL, NULL}},
  {"GT-7200", 0x1D, 0x01, {"Perfection 1250", "GT-7200"}},
  {"GT-8200", 0x1F, 0x01, {"Perfection 1650", "GT-8200"}},
  {"GT-9700", 0x21, 0x01, {"Perfection 2450", "GT-9700"}},
  {"GT-7300", 0x23, 0x01, {"Perfection 1260", "GT-7300"}},
  {"GT-8300", 0x25, 0x01, {"Perfection 1660", "GT-8300"}},
  {"GT-9300", 0x27, 0x01, {"Perfection 2400", "GT-9300"}},
  {"GT-9800", 0x29, 0x01, {"Perfection 3200", "GT-9800"}},
  {"ES-7000H", 0x2B, 0x05, {"GT-15000", "ES-7000H"}},
  {"LP-A500", 0x51, 0x01, {NULL, NULL}},
  {"AL-CX11", 0x51, 0x01, {"AcuLaser CX11", NULL}},
  {"GT-9400", 0x32, 0x06, {"Perfection 3170", "GT-9400"}},
  {"CC-600PX", 0x2D, 0x01, {"Stylus CX5100/CX5200", "CC-600PX"}},
  {"PM-A850", 0x3A, 0x01, {"Stylus Photo RX600", "PM-A850"}},
  {"CX5400", 0x36, 0x01, {"Stylus CX5300/CX5400", NULL}},
  {"GT-X700", 0x34, 0x01, {"Perfection 4870", "GT-X700"}},
  {"RX500", 0x38, 0x01, {"Stylus Photo RX500/RX510", NULL}},
  {"PX-A650", 0x37, 0x01, {"Stylus CX6300/CX6400", NULL}},
  {"ES-10000G", 0x3F, 0x05, {NULL, NULL}},
  {"Expression10000", 0x3F, 0x05, {"Expression 10000XL", NULL}},
  {"CX4600", 0x46, 0x01, {"Stylus CX4500/CX4600", NULL}},
  {"CX6600", 0x49, 0x01, {"Stylus CX6500/CX6600", NULL}},
  {"CX3600", 0x46, 0x01, {"Stylus CX3500/CX3600/CX3650", "PX-A550"}},
  {"RX420", 0x48, 0x01, {"Stylus Photo RX420/RX425/RX430", NULL}},
  {"PM-A700", 0x48, 0x01, {NULL, NULL}},
  {"PM-A870", 0x4B, 0x01, {"Stylus Photo RX620/RX630", "PM-A870"}},
  {"GT-F500", 0x41, 0x06, {"Perfection 2480/2580", "GT-F500/F550"}},
  {"GT-F600", 0x43, 0x06, {"Perfection 4180", "GT-F600"}},
  {"PM-A900", 0x4D, 0x01, {"Stylus Photo RX700", "PM-A900"}},
  {"GT-X800", 0x4F, 0x01, {"Perfection 4990", "GT-X800"}},
  {"GT-X750", 0x54, 0x01, {"Perfection 4490", "GT-X750"}},
  {"LP-M5500", 0x56, 0x05, {NULL, NULL}},
  {"LP-M5600", 0x78, 0x05, {NULL, "LP-M5600"}},
  {"GT-F520", 0x52, 0x01, {"Perfection 3490/3590", "GT-F520/F570"}},
  {"CX3800", 0x57, 0x01, {"Stylus CX3700/CX3800/DX3800", NULL}},
  {"CX7800", 0x5B, 0x01, {"Stylus CX7700/CX7800", NULL}},
  {"PM-A750", 0x5D, 0x01, {"Stylus Photo RX520/RX530", "PM-A750"}},
  {"CX4800", 0x58, 0x01, {"Stylus CX4700/CX4800/DX4800", "PX-A650"}},
  {"CX4200", 0x59, 0x01, {"Stylus CX4100/CX4200/DX4200", NULL}},
  {"PM-A950", 0x61, 0x01, {NULL, "PM-A950"}},
  {"PM-A890", 0x5F, 0x01, {"Stylus Photo RX640/RX650", "PM-A890"}},
  {"GT-X900", 0x63, 0x01, {"Perfection V700/V750", "GT-X900"}},
  {"CX4000", 0x6B, 0x01, {"Stylus CX3900/DX4000", "PX-A620"}},
  {"CX3000v", 0x6A, 0x01, {"Stylus CX2800/CX2900/ME 200", NULL}},
  {"ES-H300", 0x65, 0x01, {"GT-2500", "ES-H300"}},
  {"CX6000", 0x6C, 0x01, {"Stylus CX5900/CX6000/DX6000", "PX-A720"}},
  {"PM-A820", 0x70, 0x01, {"Stylus Photo RX560/RX580/RX590", "PM-A820"}},
  {"PM-A920", 0x71, 0x01, {NULL, "PM-A920"}},
  {"PM-A970", 0x73, 0x01, {NULL, "PM-A970"}},
  {"PM-T990", 0x75, 0x01, {NULL, "PM-T990"}},
  {"CX5000", 0x77, 0x01, {"Stylus CX4900/CX5000/DX5000", NULL}},
  {"GT-S600", 0x66, 0x01, {"Perfection V10/V100", "GT-S600/GT-F650"}},
  {"GT-F700", 0x68, 0x01, {"Perfection V350", "GT-F700"}},
  {"AL-CX21", 0x79, 0x01, {"AcuLaser CX21", NULL}},
  {"GT-F670", 0x7A, 0x01, {"Perfection V200", "GT-F670"}},
  {"GT-X770", 0x7C, 0x06, {"Perfection V500", "GT-X770"}},
  {"CX4400", 0x7E, 0x01, {"Stylus CX4300/CX4400/CX5500/CX5600/DX4400", NULL}},
  {"CX7400", 0x7F, 0x01, {"Stylus CX7300/CX7400/DX7400", "PX-A640"}},
  {"CX8400", 0x80, 0x01, {"Stylus CX8300/CX8400/DX8400", "PX-A740"}},
  {"CX9400Fax", 0x81, 0x01, {"Stylus CX9300F/CX9400Fax/DX9400F", "PX-FA700"}},
  {"PM-T960", 0x82, 0x01, {NULL, "PM-T960"}},
  {"PM-A940", 0x84, 0x01, {"Stylus Photo RX680/RX685/RX690", "PM-A940"}},
  {"PM-A840", 0x85, 0x01, {"Stylus Photo RX585/RX595/RX610", "PM-A840/PM-A840S"}},
  {"GT-D1000", 0x86, 0x01, {"GT-1500", "GT-D1000"}},
  {"GT-X970", 0x87, 0x01, {NULL, NULL}},
  {"LP-M5000", 0x97, 0x01, {NULL, NULL}},
  {"LP-M6000", 0x89, 0x01, {NULL, NULL}},
  {"ES-H7200", 0x8A, 0x05, {NULL, NULL}},
  {"GT-20000", 0x8A, 0x05, {NULL, NULL}},
  {"NX200", 0x8D, 0x01, {"Stylus NX200/SX200/TX200", NULL}},
  {"NX400", 0x8E, 0x01, {"Stylus NX400/SX400/TX400", "PX-501A"}},
  {"NX100", 0x93, 0x01, {"Stylus NX100/SX100/TX100/ME 300", "PX-401A"}},
  {"NX300", 0x8F, 0x01, {"Stylus BX300F/TX300F/NX300/ME Office 600F", NULL}},
  {"WorkForce 600", 0x90, 0x01, {"Stylus BX600FW/SX600FW/TX600FW/ME Office 700FW/WorkForce 600", "PX-601F"}},
  {"Artisan 800", 0x91, 0x01, {"Stylus Photo PX800FW/TX800FW/Artisan 800", "EP-901A/EP-901F"}},
  {"Artisan 700", 0x92, 0x01, {"Stylus Photo PX700W/TX700W/Artisan 700", "EP-801A"}},
  {"WorkForce 500", 0x96, 0x01, {NULL, NULL}},
  {"GT-F720", 0x8B, 0x01, {"Perfection V300", "GT-F720"}},
  {"GT-S620", 0x8B, 0x01, {"Perfection V30", "GT-S620"}},
  {"GT-S50", 0x00, 0x01, {"GT-S50", NULL}},
  {"GT-S80", 0x00, 0x01, {"GT-S80", "ES-D400"}},
  {"PID 0851", 0x98, 0x01, {"Stylus NX410/SX410/TX410 Series", NULL}},
  {"PID 084D", 0x99, 0x01, {"Stylus NX110/SX110/TX110 Series", "PX-402A"}},
  {"PID 084F", 0x9A, 0x01, {"Stylus NX210/SX210/TX210/ME OFFICE 510 Series", NULL}},
  {"PID 0854", 0x9B, 0x01, {"Stylus Office TX510FN/BX310FN/WorkForce 310/ME OFFICE 650FN Series", NULL}},
  {"PID 0856", 0x9C, 0x01, {"Stylus NX510/SX510W/TX550W Series", "PX-502A"}},
  {"PID 0855", 0x9D, 0x01, {"Stylus Office TX610FW/BX610FW/SX610FW/WorkForce 610 Series", "PX-602F"}},
  {"PID 0850", 0x9E, 0x01, {"Stylus Photo PX650/TX650 Series", "EP-702A"}},
  {"PID 0852", 0xA0, 0x01, {"Stylus Photo TX710W/PX710W/Artisan 710 Series", "EP-802A"}},
  {"PID 0853", 0x9F, 0x01, {"Stylus Photo PX810FW/Artisan 810 Series", "EP-902A"}},
  {"GT-X820", 0xA1, 0x01, {"Perfection V600 Photo", "GT-X820"}},

  /* array terminator */
  {NULL, 0x00, 0x00, {NULL, NULL}},
};


#define ILLEGAL_CMD   0xFF

static struct EpsonScanCommand scan_command[] = {
  {0x01, 0, 0, ILLEGAL_CMD},
  {0x02, ILLEGAL_CMD, 0, ILLEGAL_CMD},
  {0x03, 0, ILLEGAL_CMD, ILLEGAL_CMD},
  {0x04, ILLEGAL_CMD, ILLEGAL_CMD, ILLEGAL_CMD},
  {0x05, 0, 0x19, ILLEGAL_CMD},
  {0x06, 0, 0, '\f'},
};


#include "utils.h"

/* Not all scanners report the model name that is printed on the case.
   Here we try to adjust for that by second guessing the name received
   from the hardware with the aid of some environmental data.
   WARNING: this is not and can not be foolproof.

   The caller gets to manage the memory occupied by the string that is
   returned.
 */

static const char *generic_epson_scanner_name = "(unknown model)";

char *
second_guess_model_name (const char *fw_name)
{
  const scanner_data_t *data = scanner_data;
  char *result;

  struct tm *ptr = NULL;

  {				/* input validation */
    if (!fw_name || 0 == strlen (fw_name))
      return strdup (generic_epson_scanner_name);
  }

  while (data->fw_name != NULL	/* end of array */
	 && (0 != strcmp_c (data->fw_name, fw_name)))
  {
    ++data;
  }

  if (data->fw_name == NULL)	/* not a known scanner */
  {
    return strdup (fw_name);	/* use whatever we got */
  }

  /* Try to figure the name that is most likely to be on the scanner
     case.  We are using a fairly simple algorithm at the moment and
     welcome your patches ;-)  */

  result = (data->name.overseas ? data->name.overseas : data->fw_name);	/* fall back */

  {				/* get an idea of where we are */
    time_t lt = time (NULL);
    ptr = localtime (&lt);
    require (ptr);
  }

  if (data->name.japan && 0 == strncmp_c ("JST", ptr->tm_zone, 3))
  {
    result = strdup (data->name.japan);
  }
  else
  {
    result = strdup (result);
  }

  return result;
}

/*! Returns a copy of the F/W name of the device on the other end of a channel.
 */
char *
get_fw_name (channel *ch)
{
  SANE_Status status = SANE_STATUS_GOOD;
  const byte cmd[] = { ESC, 'f' };
  char *fw_name = NULL;

  if (!ch) return NULL;

  channel_send (ch, cmd, num_of (cmd), &status);
  if (SANE_STATUS_GOOD == status)
    {
      byte info[4];

      channel_recv (ch, info, num_of (info), &status);
      if (SANE_STATUS_GOOD == status)
        {
          char reply[42+1];   /* extra byte to enable string
                                 termination */
          channel_recv (ch, reply, num_of (reply) - 1, &status);
          if (SANE_STATUS_GOOD == status)
            {
              char *lc_ctype = setlocale (LC_CTYPE, "C");
              size_t n = num_of (reply);

              do
                {
                  --n, reply[n] = '\0';
                }
              while (26 < n && (isspace (reply[n-1]) || '\0' == reply[n-1]));
                                             /* workaround for non-compliant
                                                interpreter firmware names that
                                                are padded with NULs and spaces
                                              */
              fw_name = strdup (reply + 26);
              setlocale (LC_CTYPE, lc_ctype);
            }
        }
    }
  if (SANE_STATUS_GOOD != status)
    err_minor ("%s", sane_strstatus (status));

  return fw_name;
}

void
custom_cmd_settings (device *hw)
{
  const scanner_data_t *data = scanner_data;

  if (!hw) return;

  while (data->fw_name != NULL  /* end of array */
         && (0 != strcmp_c (data->fw_name, hw->fw_name)))
    {
      ++data;
    }

  {
    /* find an appropriate colour correction profile */
    int i = num_of (_epson_scan_hard);
    while (--i && data->profile_ID != epson_scan_hard[i].modelID)
      ;

    hw->scan_hard = &epson_scan_hard[i];

    if (data->command_ID)	/* customize command settings */
    {
      int id = data->command_ID - 1;	/* adjust for offset */
      require (id >= 0);
      require ((unsigned) id < num_of (scan_command));

      if (ILLEGAL_CMD != scan_command[id].set_focus_position)
      {
	hw->cmd->set_focus_position = scan_command[id].set_focus_position;
      }

      if (ILLEGAL_CMD != scan_command[id].feed)
      {
	hw->cmd->feed = scan_command[id].feed;
      }

      if (ILLEGAL_CMD != scan_command[id].eject)
      {
	hw->cmd->eject = scan_command[id].eject;
      }
    }
  }

}


static SANE_Bool
_is_listed (const char *needle, const char **haystack)
{
  if (!needle || !haystack) return SANE_FALSE;

  while (*haystack)
    {
      if (0 == strcmp_c (needle, *haystack))
        return SANE_TRUE;
      ++haystack;
    }
  return SANE_FALSE;
}


SANE_Bool
adf_needs_manual_centering (const device *hw)
{
  const char *fw_names[] = {
    "LP-M6000",
    "LP-M5000",
    "ES-H300",
    "CX9400Fax",
    NULL
  };

  require (using (hw, adf));
  return _is_listed (hw->fw_name, fw_names);
}


SANE_Bool
adf_has_auto_form_feed (const device *hw)
{
  const char *fw_names[] = {
    "LP-M6000",
    "LP-M5000",
    NULL
  };

  require (using (hw, adf));
  return _is_listed (hw->fw_name, fw_names);
}

SANE_Bool
adf_early_paper_end_kills_scan (const device *hw)
{
  const char *fw_names[] = {
    "ES-10000G",
    "ES-7000H",
    "ES-H7200",
    "Expression10000",
    "GT-20000",
    NULL,
  };

  require (using (hw, adf));
  return _is_listed (hw->fw_name, fw_names);
}

SANE_Bool
push_button_is_black_listed (const device *hw)
{
  const char *fw_names[] = {
    "LP-M6000",
    "LP-M5000",
    NULL,
  };

  // whitelist of scanners that support push button via the network
  const char *fw_names_net[] = {
    "ES-H7200",
    "GT-20000",
    NULL,
  };


  return _is_listed (hw->fw_name, fw_names)
         || (CHAN_NET == hw->channel->type &&
             !_is_listed (hw->fw_name, fw_names_net));
}

SANE_Int
large_res_kills_adf_scan (const device *hw)
{
  const char *fw_names1[] = {
    "ES-H300",
    "LP-M6000",
    "LP-M5000",
    NULL,
  };

  const char *fw_names2[] = {
    "NX300",
    "WorkForce 600",
    "Artisan 800",
    NULL,
  };

  require (hw->adf);
  if (_is_listed (hw->fw_name, fw_names1)) return 600;
  if (_is_listed (hw->fw_name, fw_names2)) return 1200;

  return 0;
}
SANE_Bool
zoom_kills_adf_scan (const device *hw)
{
  const char *fw_names[] = {
    "LP-M6000",
    "LP-M5000",
    NULL,
  };

  require (hw->adf);
  return _is_listed (hw->fw_name, fw_names);
}

SANE_Bool
adf_duplex_direction_matches (const device *hw)
{
  const char *fw_names[] = {
    "GT-S80",
    "GT-S50",
    NULL,
  };

  return hw->adf && _is_listed (hw->fw_name, fw_names);
}
