/*  message.h -- consistent error, progress and debugging feedback
 *  Copyright (C) 2008, 2009  SEIKO EPSON CORPORATION
 *
 *  License: GPLv2+|iscan
 *  Authors: AVASYS CORPORATION
 *
 *  This file is part of the SANE backend distributed with Image Scan!
 *
 *  Image Scan!'s SANE backend is free software.
 *  You can redistribute it and/or modify it under the terms of the GNU
 *  General Public License as published by the Free Software Foundation;
 *  either version 2 of the License or at your option any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *  You ought to have received a copy of the GNU General Public License
 *  along with this package.  If not, see <http://www.gnu.org/licenses/>.
 *
 *
 *  Linking Image Scan!'s SANE backend statically or dynamically with
 *  other modules is making a combined work based on this SANE backend.
 *  Thus, the terms and conditions of the GNU General Public License
 *  cover the whole combination.
 *
 *  As a special exception, the copyright holders of Image Scan!'s SANE
 *  backend give you permission to link Image Scan!'s SANE backend with
 *  SANE frontends that communicate with Image Scan!'s SANE backend
 *  solely through the SANE Application Programming Interface,
 *  regardless of the license terms of these SANE frontends, and to
 *  copy and distribute the resulting combined work under terms of your
 *  choice, provided that every copy of the combined work is
 *  accompanied by a complete copy of the source code of Image Scan!'s
 *  SANE backend (the version of Image Scan!'s SANE backend used to
 *  produce the combined work), being distributed under the terms of
 *  the GNU General Public License plus this exception.  An independent
 *  module is a module which is not derived from or based on Image
 *  Scan!'s SANE backend.
 *
 *  As a special exception, the copyright holders of Image Scan!'s SANE
 *  backend give you permission to link Image Scan!'s SANE backend with
 *  independent modules that communicate with Image Scan!'s SANE
 *  backend solely through the "Interpreter" interface, regardless of
 *  the license terms of these independent modules, and to copy and
 *  distribute the resulting combined work under terms of your choice,
 *  provided that every copy of the combined work is accompanied by a
 *  complete copy of the source code of Image Scan!'s SANE backend (the
 *  version of Image Scan!'s SANE backend used to produce the combined
 *  work), being distributed under the terms of the GNU General Public
 *  License plus this exception.  An independent module is a module
 *  which is not derived from or based on Image Scan!'s SANE backend.
 *
 *  Note that people who make modified versions of Image Scan!'s SANE
 *  backend are not obligated to grant special exceptions for their
 *  modified versions; it is their choice whether to do so.  The GNU
 *  General Public License gives permission to release a modified
 *  version without this exception; this exception also makes it
 *  possible to release a modified version which carries forward this
 *  exception.
 */


#ifndef message_h_included
#define message_h_included

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>

#ifdef __cplusplus
extern "C"
{
#endif


  typedef enum
    {
      ERR_FATAL = (1 << 0),
      ERR_MAJOR = (1 << 1),
      ERR_MINOR = (1 << 2),

      LOG_INFO  = (1 << 3),
      LOG_CALL  = (1 << 4),
      LOG_DATA  = (1 << 5),

      DBG_CMD   = (1 << 6),
      DBG_HEX   = (1 << 7),
      DBG_IMG   = (1 << 8),
    }
    msg_level_type;


  extern unsigned long msg_level;


#define FMT_FILE __FILE__ ":%d: "
#define FMT_LINE , __LINE__
#define FMT_BACKEND FMT_FILE "[epkowa]"

#define FMT_FATAL  FMT_BACKEND "[F] "
#define FMT_MAJOR  FMT_BACKEND "[M] "
#define FMT_MINOR  FMT_BACKEND "[m] "
#define FMT_INFO   FMT_BACKEND "{I} "
#define FMT_CALL   FMT_BACKEND "{C} " "%s "
#define FMT_DATA   FMT_BACKEND "{D} "
#define FMT_CMD    FMT_BACKEND "(e) "
#define FMT_HEX    "[epkowa]" "(x) "
#define FMT_IMG    "[epkowa]" "(i) "


#define err_fatal(fmt,arg...)                   \
  if (ERR_FATAL <= msg_level)                   \
    fprintf (stderr, FMT_FATAL fmt "\n" FMT_LINE, ## arg)

#define err_major(fmt,arg...)                   \
  if (ERR_MAJOR <= msg_level)                   \
    fprintf (stderr, FMT_MAJOR fmt "\n" FMT_LINE, ## arg)

#define err_minor(fmt,arg...)                   \
  if (ERR_MINOR <= msg_level)                   \
    fprintf (stderr, FMT_MINOR fmt "\n" FMT_LINE, ## arg)

#define log_info(fmt,arg...)                    \
  if (LOG_INFO <= msg_level)                    \
    fprintf (stderr, FMT_INFO fmt "\n" FMT_LINE, ## arg)

#define log_call(fmt,arg...)                    \
  if (LOG_CALL <= msg_level)                    \
    fprintf (stderr, FMT_CALL fmt "\n" FMT_LINE, __func__, ## arg)

#define log_data(fmt,arg...)                    \
  if (LOG_DATA <= msg_level)                    \
    fprintf (stderr, FMT_DATA fmt "\n" FMT_LINE, ## arg)

#define dbg_cmd(buf,sz)                         \
  if (DBG_CMD <= msg_level)                     \
    fprintf (stderr, FMT_CMD fmt "\n" FMT_LINE, ## arg)

#define dbg_hex(buf,sz)                         \
  if (DBG_HEX <= msg_level)                     \
    msg_dump (FMT_HEX, buf, sz);

#define dbg_img(buf,sz)                         \
  if (DBG_IMG <= msg_level)                     \
    msg_dump (FMT_IMG, buf, sz);


  void msg_init (void);
  void msg_dump (const char *, const void *, size_t);


#ifdef __cplusplus
}       /* extern "C" */
#endif

#endif  /* !defined (message_h_included) */
